"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the process function.

Routine listings
----------------
process(func, namespace={}, args_list=None, kwargs_list=None, maxtasks=None)
    Map multiple function calls to multiple processors.

See Also
--------
magni.utils.multiprocessing.config : Configuration options.

"""

from __future__ import division
import multiprocessing as mp
import traceback
import types

# disable mkl multiprocessing to avoid conflicts with manual multiprocessing
try:
    import mkl
    _get_num_threads = mkl.get_max_threads
    _set_num_threads = mkl.set_num_threads
except ImportError:
    _get_num_threads = lambda: 0
    _set_num_threads = lambda n: 0

from magni.utils.multiprocessing import config as _config
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_process(func, namespace, args_list, kwargs_list, maxtasks):
    """
    Validation of the `process` function.

    See Also
    --------
    process : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(func, 'func', [{'type': types.FunctionType}])
    _validate(namespace, 'namespace', [{'type': dict}])
    requirement = {'type_in': [list, tuple]}
    _validate(args_list, 'args_list', [requirement, requirement], True)
    _validate(kwargs_list, 'kwargs_list', [requirement, {'type': dict}], True)

    if args_list is None and kwargs_list is None:
        raise ValueError('args_list and kwargs_list cannot both be None.')
    elif args_list is not None and kwargs_list is not None:
        if len(args_list) != len(kwargs_list):
            raise ValueError('len(args_list) must equal len(kwargs_list).')

    _validate(maxtasks, 'maxtasks', {'type': int}, True)


def process(func, namespace={}, args_list=None, kwargs_list=None,
            maxtasks=None):
    """
    Map multiple function calls to multiple processors.

    For each entry in args_list and kwargs_list, a task is formed which is used
    for a function call of the type `func(*args, **kwargs)`.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    namespace : dict, optional
        A dict whose keys and values should be globally available in func (the
        default is an empty dict).
    args_list : list or tuple, optional
        A sequence of argument lists for the function calls (the default is
        None, which implies that no arguments are used in the calls).
    kwargs_list : list or tuple, optional
        A sequence of keyword argument dicts for the function calls (the
        default is None, which implies that no keyword arguments are used in
        the calls).
    maxtasks : int, optional
        The maximum number of tasks of a process before it is replaced by a new
        process (the default is None, which implies that processes are not
        replaced).

    Returns
    -------
    results : list
        A list with the results from the function calls.

    See Also
    --------
    magni.utils.multiprocessing.config : Configuration options.

    Notes
    -----
    If the `workers` configuration option is equal to 0, map is used.
    Otherwise, the map functionality of a multiprocessing worker pool is used.

    Reasons for using this function over map or standard multiprocessing:

    - Simplicity of the code over standard multiprocessing.
    - Simplicity in switching between single- and multiprocessing.
    - The use of both arguments and keyword arguments in the function calls.
    - The reporting of exceptions before termination.
    - The possibility of terminating multiprocessing with a single interrupt.

    Examples
    --------
    An example of how to use args_list, and kwargs_list:

    >>> from magni.utils.multiprocessing._processing import process
    >>> def calculate(a, b, op='+'):
    ...     if op == '+':
    ...         return a + b
    ...     elif op == '-':
    ...         return a - b
    ...
    >>> args_list = [[5, 7], [9, 3]]
    >>> kwargs_list = [{'op': '+'}, {'op': '-'}]
    >>> process(calculate, args_list=args_list, kwargs_list=kwargs_list)
    [12, 6]

    """

    _validate_process(func, namespace, args_list, kwargs_list, maxtasks)

    if args_list is None:
        args_list = [[] for dct in kwargs_list]

    if kwargs_list is None:
        kwargs_list = [{} for lst in args_list]

    tasks = [func for lst in args_list]
    tasks = list(zip(tasks, args_list, kwargs_list))

    if _config.get('workers') == 0:
        _process_init(func, namespace)
        results = list(map(_process_worker, tasks))
    else:
        try:
            num_threads = _get_num_threads()
            _set_num_threads(1)
            workers = mp.Pool(_config.get('workers'), _process_init,
                              (func, namespace), maxtasks)
            results = workers.map(_process_worker, tasks, chunksize=1)
        finally:
            workers.close()
            workers.join()
            _set_num_threads(num_threads)

    return results


def _process_init(func, namespace):
    """
    Initialise the process by making global variables available to it.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    namespace : dict
        A dict whose keys and values should be globally available in func.

    """

    func.__globals__.update(namespace)


def _process_worker(fak_tuple):
    """
    Unpack and map a task to the function.

    Parameters
    ----------
    fak_tuple : tuple
        A tuple (func, args, kwargs) containing the parameters listed below.
    func : function
        A function handle to the function which the calls should be mapped to.
    args : list or tuple
        The sequence of arguments that should be unpacked and passed.
    kwargs : list or tuple
        The sequence of keyword arguments that should be unpacked and passed.

    Notes
    -----
    If an exception is raised in `func`, the stacktrace of that exception is
    printed since the exception is otherwise silenced until every task has been
    executed when using multiple workers.

    Also, a workaround has been implemented to allow KeyboardInterrupts to
    interrupt the current tasks and all remaining tasks. This is done by
    setting a global variable, when catching a KeyboardInterrupt, which is
    checked for every call.

    """

    func, args, kwargs = fak_tuple

    if 'interrupted' not in _process_worker.__globals__:
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            _process_worker.__globals__['interrupted'] = True
        except BaseException:
            print(traceback.format_exc())
            raise RuntimeError('An exception has occured.')



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality related linear transformations.

Routine listings
----------------
dct2(x, m, n)
    2D discrete cosine transform.
idct2(x, m, n)
    2D inverse discrete cosine tranform.
dft2(x, m, n)
    2D discrete Fourier transform.
idft2(x, m, n)
    2D inverse discrete Fourier transform.

"""

from __future__ import division

import numpy as np
import scipy.fftpack

from magni.imaging._util import mat2vec as _mat2vec
from magni.imaging._util import vec2mat as _vec2mat
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_transform(x, m, n):
    """
    Validatate a 2D transform.

    See also
    --------
    magni.utils.validation.validate : Validation.

    """

    _validate(m, 'm', {'type': int, 'min': 1})
    _validate(n, 'n', {'type': int, 'min': 1})
    _validate_ndarray(x, 'x', {'shape': (m * n, 1)})


def dct2(x, m, n):
    """
    Apply the 2D Discrete Cosine Transform (DCT) to `x`.

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    m : int
        Number of rows in the associated matrix.
    n : int
        Number of columns in the associated matrix.

    Returns
    -------
    ndarray
        A m*n x 1 vector of coefficients scaled such that x = idct2(dct2(x)).

    See Also
    --------
    scipy.fftpack.dct : 1D DCT

    """

    _validate_transform(x, m, n)

    # 2D DCT using the seperability property of the 1D DCT.
    # http://stackoverflow.com/questions/14325795/scipys-fftpack-dct-and-idct
    # Including the reshape operation, the full transform is:
    # dct(dct(x.reshape(n, m).T).T).T.T.reshape(m * n, 1)

    result = _vec2mat(x, (m, n))
    result = scipy.fftpack.dct(result, norm='ortho').T
    result = scipy.fftpack.dct(result, norm='ortho').T
    return _mat2vec(result)


def idct2(x, m, n):
    """
    Apply the 2D Inverse Discrete Cosine Transform (iDCT) to `x`.

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    m : int
        Number of rows in the associated matrix.
    n : int
        Number of columns in the associated matrix.

    Returns
    -------
    ndarray
        A m*n x 1 vector of coefficients scaled such that x = dct2(idct2(x)).

    See Also
    --------
    scipy.fftpack.idct : 1D inverse DCT

    """

    _validate_transform(x, m, n)

    result = _vec2mat(x, (m, n))
    result = scipy.fftpack.idct(result, norm='ortho').T
    result = scipy.fftpack.idct(result, norm='ortho').T
    return _mat2vec(result)


def dft2(x, m, n):
    """
    Apply the 2D Discrete Fourier Transform (DFT) to `x`.

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    m : int
        Number of rows in the associated matrix.
    n : int
        Number of columns in the associated matrix.

    Returns
    -------
    ndarray
        A m*n x 1 vector of coefficients scaled such that x = dft2(idft2(x)).

    See Also
    --------
    numpy.fft.fft2 : The underlying 2D FFT used to compute the 2D DFT.

    """

    _validate_transform(x, m, n)

    return _mat2vec(np.fft.fft2(_vec2mat(x, (m, n))))


def idft2(x, m, n):
    """
    Apply the 2D Inverse Discrete Fourier Transform (iDFT) to `x`.

    `x` is assumed to be the column vector resulting from stacking the columns
    of the associated matrix which the transform is to be taken on.

    Parameters
    ----------
    x : ndarray
        The m*n x 1 vector representing the associated column stacked matrix.
    m : int
        Number of rows in the associated matrix.
    n : int
        Number of columns in the associated matrix.

    Returns
    -------
    ndarray
        A m*n x 1 vector of coefficients scaled such that x = idft2(dft2(x)).

    See Also
    --------
    numpy.fft.ifft2 : The underlying 2D iFFT used to compute the 2D iDFT.

    """

    _validate_transform(x, m, n)

    output = _mat2vec(np.fft.ifft2(_vec2mat(x, (m, n))))

    if np.abs(output.imag).sum() == 0:
        output = output.real

    return output



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.sl0`
subpackage.

Routine listings
----------------
get(key=None)
    Get the value of one or more configuration options.
set(dictionary={}, \*\*kwargs)
    Set the value of one or more configuration options.

See Also
--------
magni.utils.config : The Configger class used.

Notes
-----
This module instantiates the `Configger` class provided by `magni.utils.config`
and assigns handles for the `get` and `set` methods of that class instance. The
configuration options are the following:

epsilon : float
    The precision parameter used in centering (the default is 0.01).
L : float
    The maximum number of gradient descent iterations for each sigma (the
    default is 2.0).
L_update : float
    The additive/multiplicative update of L when it is not fixed (the default
    is 2.0).
mu : float
    The relative step-size used in gradient descent iteration (the default is
    1.0).
mu_start : float
    The relative step-size used in gradient descent iteration for the first
    iterations (the default is 0.001).
mu_end :float
    The relative step-size used in gradient descent iteration for the last
    iterations (the default is 1.5).
precision_float : dtype
    The floating point precision used for the computations (the default is
    float64).
seq_algorithm : {'mod', 'std'}
    The sequential implementation of SL0 used (the default is 'mod').
sigma_update : float
    The constant used in the geometric sequence of sigmas (the default is 0.7).
sigma_min : float
    The minimum value of std. dev. in Gaussian l0 approx (the default 0.01).

"""

from __future__ import division

import numpy as np

from magni.utils.config import Configger as _Configger


_configger = _Configger(
    {'sigma_update': 0.7, 'sigma_min': 0.01, 'L': 2.0, 'L_update': 2.0,
     'mu': 1.0, 'mu_start': 0.001, 'mu_end': 1.5, 'epsilon': 1e-2,
     'precision_float': np.float64, 'algorithm': 'mod'},
    {'sigma_update': {'type': float},
     'sigma_min': {'type': float},
     'L': {'type': float},
     'L_update': {'type': float},
     'mu': {'type': float},
     'mu_start': {'type': float},
     'mu_end': {'type': float},
     'epsilon': {'type': float},
     'precision_float': {'type': type},
     'algorithm': {'type': str, 'val_in': ['std', 'mod']}})

set = _configger.set


def get(key=None):
    """
    Get the value of one or more configuration options.

    This function wraps 'Configger.get' in order to convert any float options
    to the specified precision before returning the option(s).

    See Also
    --------
    magni.utils.config.Configger.get : The wrapped function.

    """

    value = _configger.get(key)
    convert = _configger.get('precision_float')

    if isinstance(value, dict):
        for key, val in value.items():
            if isinstance(val, float):
                value[key] = convert(val)
    elif isinstance(value, float):
        value = convert(value)

    return value



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the actual reconstruction algorithm.

Routine listings
----------------
run(y, A)
    Run the IHT reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.iht.config : Configuration options.

Notes
-----
The IHT reconstruction algorithm is described in [1]_.

References
----------
.. [1] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
   Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in Signal
   Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.

"""

from __future__ import division

import numpy as np
import scipy.stats

from magni.cs.reconstruction.iht import config as _config
from magni.utils.matrices import Matrix as _Matrix
from magni.utils.matrices import MatrixCollection as _MatrixC
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_ndarray as _validate_ndarray


@_decorate_validation
def _validate_run(y, A):
    """
    Validate the `run` function.

    See Also
    --------
    run : The validated function.
    magni.utils.validation.validate : Validation.

    """

    try:
        _validate_ndarray(A, 'A')
    except TypeError:
        if not isinstance(A, _Matrix) and not isinstance(A, _MatrixC):
            raise TypeError('A must be a matrix.')

    _validate_ndarray(y, 'y', {'shape': (A.shape[0], 1)})


def run(y, A):
    """
    Run the IHT reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.

    See Also
    --------
    _calculate_far : Optimal False Acceptance Rate calculation.
    _normalise : Matrix normalisation.

    Notes
    -----
    In each iteration, the threshold is robustly calculated as a fixed multiple
    of the standard deviation of the calculated correlations. The fixed
    multiple is based on the False Acceptance Rate (FAR) assuming a Gaussian
    distribution of the correlations.

    The algorithm terminates after a fixed number of iterations or if the ratio
    between the 2-norm of the residual and the 2-norm of the measurements falls
    below the specified `tolerance`.

    Examples
    --------
    For example, recovering a vector from random measurements

    >>> from magni.cs.reconstruction.iht._original import run
    >>> np.random.seed(seed=6021)
    >>> A = 1 / np.sqrt(80) * np.random.randn(80, 200)
    >>> x = np.zeros((200, 1))
    >>> x[:10] = 1
    >>> y = A.dot(x)
    >>> x_hat = run(y, A)
    >>> x_hat[:12]
    array([[ 0.99836297],
           [ 1.00029086],
           [ 0.99760224],
           [ 0.99927175],
           [ 0.99899124],
           [ 0.99899434],
           [ 0.9987368 ],
           [ 0.99801849],
           [ 1.00059408],
           [ 0.9983772 ],
           [ 0.        ],
           [ 0.        ]])
    >>> (np.abs(x_hat) > 1e-2).sum()
    10

    """

    _validate_run(y, A)

    _param = _config.get()
    convert = _param['precision_float']
    kappa = _param['kappa']
    tol = _param['tolerance']
    far = _calculate_far(A.shape[0] / A.shape[1])
    Lambda = convert(scipy.stats.norm.ppf(1 - far / 2))
    stdQ1 = convert(scipy.stats.norm.ppf(1 - 0.25))
    k = int(_param['threshold_rho'] * A.shape[0])

    x = np.zeros((A.shape[1], 1), dtype=convert)
    r = y.copy()

    for it in range(_param['iterations']):
        c = A.T.dot(r)
        x = x + kappa * c

        if _param['threshold'] == 'far':
            thres = (kappa * Lambda * convert(np.median(np.abs(c.ravel())))
                     / stdQ1)
        elif _param['threshold'] == 'oracle':
            if k == 0:
                thres = np.abs(x.ravel()).max() + 1
            elif k == A.shape[0]:
                thres = 0
            else:
                thres = np.sort(np.abs(x.ravel()))[-(k + 1)]

        x[np.abs(x) <= thres] = 0
        r = y - A.dot(x)

        if np.linalg.norm(r) < tol * np.linalg.norm(y):
            break

    return x


def _calculate_far(delta):
    """
    Calculate the optimal False Acceptance Rate for a given indeterminacy.

    Parameters
    ----------
    delta : float
        The indeterminacy, m / n, of a system of equations of size m x n.

    Returns
    -------
    FAR : float
        The optimal False Acceptance Rate for the given indeterminacy.

    Notes
    -----
    The optimal False Acceptance Rate to be used in connection with the
    interference heuristic presented in the paper "Optimally Tuned Iterative
    Reconstruction Algorithms for Compressed Sensing" [2]_ is calculated from
    a set of optimal values presented in the same paper. The calculated value
    is found from a linear interpolation or extrapolation on the known set of
    optimal values.

    References
    ----------
    .. [2] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
       Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in
       Signal Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.

    """

    # Known optimal values (x - indeterminacy / y - FAR)
    x = [0.05, 0.11, 0.21, 0.41, 0.50, 0.60, 0.70, 0.80, 0.93]
    y = [0.0015, 0.002, 0.004, 0.011, 0.015, 0.02, 0.027, 0.035, 0.043]

    for i in range(len(x) - 1):
        if delta <= x[i + 1]:
            break

    return y[i] + (delta - x[i]) * (y[i + 1] - y[i]) / (x[i + 1] - x[i])



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of generic reconstruction algorithms.

Each subpackage provides a family of generic reconstruction algorithms. Thus
each subpackage has a config module and a run function which provide the
interface of the given family of reconstruction algorithms.

Routine listings
----------------
iht
    Subpackage providing an implementation of Iterative Hard Thresholding (IHT)
sl0
    Subpackage providing implementations of Smoothed l0 Norm (SL0).

"""

from magni.cs.reconstruction import iht
from magni.cs.reconstruction import sl0



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of Smoothed l0 Norm SL0).

The implementations provided are the original SL0 reconstruction algorithm and
a modified SL0 reconstruction algorithm. The algorithm used depends on the
'algorithm' configuration option: 'std' refers to the original algorithm while
'mod' (default) refers to the modified algorithm.

Routine listings
----------------
config
    Module providing configuration options for this subpackage.
run(y, A)
    Run the specified SL0 reconstruction algorithm.

Notes
-----
See `_util` for documentation of `run`.

The original SL0 reconstruction algorithm by Mohimani et. al is described in
[1]_ whereas the constraint elimiation intepretation of the original SL0
algorithm by Cui et. al. is described in [2]_. The modified SL0 reconstruction
algorithm by Oxvig et. al. is described in [3]_. Specifically, the provided
sequential implementations are:

| std : The standard SL0 algorithm
|     For delta < 0.55: Standard projection algorithm by Mohimani et. al.
|     For delta >= 0.55: Standard constraint elimination algorithm by Cui et.
      al.
| mod : The modified SL0 algorithm (the default)
|     For delta < 0.55: Modified projection algorithm
|     For delta >= 0.55: Modified constraint elimination algorithm

References
----------
.. [1] H. Mohimani, M. Babaie-Zadeh, and C. Jutten, "A Fast Approach for
   Overcomplete Sparse Decomposition Based on Smoothed l0 Norm", *IEEE
   Transactions on Signal Processing*, vol. 57, no. 1, pp. 289-301, Jan. 2009.
.. [2] Z. Cui, H. Zhang, and W. Lu, "An Improved Smoothed l0-norm Algorithm
   Based on Multiparameter Approximation Function", *in 12th IEEE International
   Conference on Communication Technology (ICCT)*, Nanjing, China, Nov. 11-14,
   2011, pp. 942-945.
.. [3] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from magni.cs.reconstruction.sl0 import config
from magni.cs.reconstruction.sl0._util import run



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input/output functionality for stored phase transitions.

Routine listings
----------------
load_phase_transition(path, label='default')
    Load the coordinates of a phase transition from a HDF5 file.

"""

from __future__ import division

import numpy as np

from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_load_phase_transition(path, label):
    """
    Validate the `load_phase_transition` function.

    See Also
    --------
    load_phase_transition : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(path, 'path', {'type': str})
    _validate(label, 'label', {'type': str})


def load_phase_transition(path, label='default'):
    """
    Load the coordinates of a phase transition from a HDF5 file.

    This function is used to load the phase transition from the output file
    generated by `magni.cs.phase_transition.determine`.

    Parameters
    ----------
    path : str
        The path of the HDF5 file where the phase transition is stored.
    label : str
        The label assigned to the phase transition (the default is 'default').

    Returns
    -------
    delta : np.ndarray
        The delta values of the phase transition points.
    rho : np.ndarray
        The rho values of the phase transition points.

    See Also
    --------
    magni.cs.phase_transition.determine : Phase transition determination.
    magni.cs.phase_transition.plotting : Phase transition plotting.

    Examples
    --------
    An example of how to use this function is provided in the `examples` folder
    in the `cs-phase_transition.ipynb` ipython notebook file.

    """

    _validate_load_phase_transition(path, label)

    with _File(path, 'r') as f:
        rho = f.get_node('/' + label + '/phase_transition')[:]
        delta = np.linspace(0, 1, len(rho) + 1)[1:]
        return (delta, rho)



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input/output functions to databases containing results from
reproducible research.

Routine listings
----------------
annotate_database(h5file)
    Function for annotating an existing HDF5 database.
read_annotations(h5file)
    Function for reading annotations in an HDF5 database.
remove_annotations(h5file)
    Function for removing annotations in an HDF5 database.

See Also
--------
magni.reproducibility._annotation.get_conda_info : Conda annotation
magni.reproducibility._annotation.get_git_revision : Git annotation
magni.reproducibility._annotation.get_platform_info : Platform annotation
magni.reproducibility._annotation.get_datetime : Date and time annotation
magni.reproducibility._annotation.get_magni_config : Magni config annotation
magni.reproducibility._annotation.get_magni_info : Magni info annotation

"""

from __future__ import division
import json

import tables

from magni.reproducibility import _annotation
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate as _validate


@_decorate_validation
def _validate_annotate_database(h5file):
    """
    Validate the `annotate_database` function.

    See Also
    --------
    annotate_database : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h5file, 'h5file', {'class': tables.file.File})


def annotate_database(h5file):
    """
    Annotate an HDF5 database with information about Magni and the platform.

    The annotation consists of a group in the root of the `h5file` having nodes
    that each provide information about Magni or the platform on which this
    function is run.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database that should be annotated.

    See Also
    --------
    magni.reproducibility._annotation.get_conda_info : Conda annotation
    magni.reproducibility._annotation.get_git_revision : Git annotation
    magni.reproducibility._annotation.get_platform_info : Platform annotation
    magni.reproducibility._annotation.get_datetime : Date and time annotation
    magni.reproducibility._annotation.get_magni_config : Magni config
        annotation
    magni.reproducibility._annotation.get_magni_info : Magni info annotation

    Notes
    -----
    The annotations of the database includes the following:

    * conda_info - Information about Continuum Anacononda install
    * git_revision - Git revision and tag of Magni
    * platform_info - Information about the current platform (system)
    * datetime - The current date and time
    * magni_config - Infomation about the current configuration of Magni
    * magni_info - Information from `help(magni)`

    Examples
    --------
    Annotate the database named 'db.hdf5':

    >>> from magni.reproducibility.io import annotate_database
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...     annotate_database(h5file)

    """

    _validate_annotate_database(h5file)

    annotations = {'conda_info': json.dumps(_annotation.get_conda_info()),
                   'git_revision': json.dumps(_annotation.get_git_revision()),
                   'platform_info': json.dumps(
                       _annotation.get_platform_info()),
                   'datetime': json.dumps(_annotation.get_datetime()),
                   'magni_config': json.dumps(_annotation.get_magni_config()),
                   'magni_info': json.dumps(_annotation.get_magni_info())}

    try:
        annotations_group = h5file.create_group('/', 'annotations')
        for annotation in annotations:
            h5file.create_array(annotations_group, annotation,
                                obj=annotations[annotation].encode())
        h5file.flush()
    except tables.NodeError:
        raise tables.NodeError('The database has already been annotated. ' +
                               'Remove the existing annotation prior to ' +
                               '(re)annotating the database.')


@_decorate_validation
def _validate_read_annotations(h5file):
    """
    Validate the `read_annotations` function.

    See Also
    --------
    read_annotations : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h5file, 'h5file', {'class': tables.file.File})


def read_annotations(h5file):
    """
    Read the annotations to an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the annotations is read.

    Returns
    -------
    annotations : dict
        The annotations read from the HDF5 database.

    Raises
    ------
    ValueError
        If the annotations to the HDF5 database does not conform to the Magni
        annotation standard.

    Notes
    -----
    The returned dict holds a key for each annotation in the database. The
    value corresponding to a given key is in itself a dict. See
    `magni.reproducibility.annotate_database` for examples of such annotations.

    Examples
    --------
    Read annotations from the database named 'db.hdf5':

    >>> from magni.reproducibility.io import read_annotations
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='r') as h5file:
    ...    annotations = read_annotations(h5file)

    """

    _validate_read_annotations(h5file)

    try:
        h5_annotations = h5file.get_node('/', name='annotations')
    except tables.NoSuchNodeError:
        raise tables.NoSuchNodeError('The database has not been annotated.')

    h5_annotations_dict = h5_annotations._v_leaves
    annotations = dict()
    try:
        for annotation in h5_annotations_dict:
            annotations[annotation] = json.loads(
                h5_annotations_dict[annotation].read().decode())
    except ValueError as e:
        raise ValueError('Unable to read the {!r} '.format(annotation) +
                         'annotation. It seems that the annotation ' +
                         'does not conform to the Magni annotation ' +
                         'standard ({!r}).'.format(e.args[0]))

    return annotations


@_decorate_validation
def _validate_remove_annotations(h5file):
    """
    Validate the `remove_annotations` function.

    See Also
    --------
    remove_annotations : The validated function.
    magni.utils.validation.validate : Validation.

    """

    _validate(h5file, 'h5file', {'class': tables.file.File})


def remove_annotations(h5file):
    """
    Remove the annotations from an HDF5 database.

    Parameters
    ----------
    h5file : tables.file.File
        The handle to the HDF5 database from which the annotations is removed.

    Examples
    --------
    Remove annotations from the database named 'db.hdf5':

    >>> from magni.reproducibility.io import remove_annotations
    >>> with magni.utils.multiprocessing.File('db.hdf5', mode='a') as h5file:
    ...    remove_annotations(h5file)

    """

    _validate_remove_annotations(h5file)

    try:
        h5file.remove_node('/', 'annotations', recursive=True)
        h5file.flush()
    except tables.NoSuchNodeError:
        pass



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing an implementation of Iterative Hard Thresholding (IHT).

Routine listings
----------------
config
    Module providing configuration options for this subpackage.
run(y, A)
    Run the IHT reconstruction algorithm.

Notes
-----
See `_original` for documentation of `run`.

The IHT reconstruction algorithm is described in [1]_.

References
----------
.. [1] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
   Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in Signal
   Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.

"""

from magni.cs.reconstruction.iht import config
from magni.cs.reconstruction.iht._original import run



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the multiprocessing subpackage.

Routine listings
----------------
get(key=None)
    Get the value of one or more configuration options.
set(dictionary={}, \*\*kwargs)
    Set the value of one or more configuration options.

See Also
--------
magni.utils.config.Configger : The Configger class used.

Notes
-----
This module instantiates the `Configger` class provided by `magni.utils.config`
and assigns handles for the `get` and `set` methods of that class instance. The
configuration options are the following:

workers : int
    The number of workers to use for multiprocessing (the default is 0, which
    implies no multiprocessing).

"""

from __future__ import division

from magni.utils.config import Configger as _Configger


_configger = _Configger({'workers': 0}, {'workers': {'type': int, 'min': 0}})
get = _configger.get
set = _configger.set



"""
..
    Copyright (c) 2014, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.iht`
subpackage.

Routine listings
----------------
get(key=None)
    Get the value of one or more configuration options.
set(dictionary={}, \*\*kwargs)
    Set the value of one or more configuration options.

See Also
--------
magni.utils.config : The Configger class used.

Notes
-----
This module instantiates the `Configger` class provided by `magni.utils.config`
and assigns handles for the `get` and `set` methods of that class instance. The
configuration options are the following:

iterations : int
    The maximum number of iterations to do (the default is 300).
kappa : float
    The relaxation parameter used in the algorithm (the default is 0.65).
precision_float : dtype
    The floating point precision used for the computations (the default is
    float64).
threshold : ['far', 'oracle']
    The method for selecting the threshold value.
threshold_rho : float
    The assumed rho value used for selecting the threshold value if using the
    oracle method.
tolerance : float
    The least acceptable ratio of residual to measurements (in 2-norm) to break
    the interations (the default is 0.001).

"""

from __future__ import division

import numpy as np

from magni.utils.config import Configger as _Configger


_configger = _Configger(
    {'kappa': 0.65, 'tolerance': 1e-3, 'iterations': 300, 'threshold': 'far',
     'threshold_rho': 0.1, 'precision_float': np.float64},
    {'kappa': {'type': float, 'min': 0},
     'tolerance': {'type': float, 'min': 0},
     'iterations': {'type': int, 'min': 1},
     'threshold': {'type': str, 'val_in': ('far', 'oracle')},
     'threshold_rho': {'type': float, 'min': 0, 'max': 1},
     'precision_float': {'type': type}})

set = _configger.set


def get(key=None):
    """
    Get the value of one or more configuration options.

    This function wraps 'Configger.get' in order to convert any float options
    to the specified precision before returning the option(s).

    See Also
    --------
    magni.utils.config.Configger.get : The wrapped function.

    """

    value = _configger.get(key)
    convert = _configger.get('precision_float')

    if isinstance(value, dict):
        for key, val in value.items():
            if isinstance(val, float):
                value[key] = convert(val)
    elif isinstance(value, float):
        value = convert(value)

    return value
